# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
import numpy as np
cimport numpy as np
from cython.parallel import prange
from libc.stdint cimport uint8_t

ctypedef uint8_t u8
ctypedef float f32

cdef f32 PRO_MIN = 1e-4
cdef f32 PRO_MAX = 2.0 - (1e-4)
cdef inline f32 _fmax(f32 a, f32 b) noexcept nogil: return a if a > b else b
cdef inline f32 _fmin(f32 a, f32 b) noexcept nogil: return a if a < b else b
cdef inline f32 _clamp1(f32 a) noexcept nogil: return _fmax(PRO_MIN, _fmin(a, PRO_MAX))


##### EMU-mem #####
# Inline functions
cdef inline f32 innerE(
		const f32* u, const f32* v, const f32 f, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f32 d = 2.0*f
		f32 e = d
	for k in range(K):
		e += u[k]*v[k]
	return _clamp1(e) - d

# Extract and center chunk (frequencies) for randomized SVD
cpdef void memCenter(
		const u8[:,::1] G, f32[:,::1] X, f32[::1] f, const Py_ssize_t M_w
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, l, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fl
	for j in prange(M, schedule='guided'):
		i = 0
		l = M_w + j
		fl = f[l]
		for b in range(B):
			byte = G[l,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					X[j,i] = <f32>g - 2.0*fl
				else:
					X[j,i] = 0.0
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Extract and center chunk (SVD) for randomized SVD
cpdef void memCenterSVD(
		const u8[:,::1] G, f32[:,::1] U, const f32[:,::1] V, f32[:,::1] X, f32[::1] f, const Py_ssize_t M_w
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		Py_ssize_t B = G.shape[1]
		Py_ssize_t K = U.shape[1]
		size_t b, i, j, l, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fl
		f32* Ul
	for j in prange(M, schedule='guided'):
		i = 0
		l = M_w + j
		fl = f[l]
		Ul = &U[l,0]
		for b in range(B):
			byte = G[l,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					X[j,i] = <f32>g - 2.0*fl
				else:
					X[j,i] = innerE(Ul, &V[i,0], fl, K)
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Extract and standardize chunk (frequencies) for randomized SVD
cpdef void memFinal(
		const u8[:,::1] G, f32[:,::1] X, f32[::1] f, f32[::1] d, const Py_ssize_t M_w
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, l, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fl, dl
	for j in prange(M, schedule='guided'):
		i = 0
		l = M_w + j
		fl = f[l]
		dl = d[l]
		for b in range(B):
			byte = G[l,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					X[j,i] = (<f32>g - 2.0*fl)*dl
				else:
					X[j,i] = 0.0
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Extract and standardize chunk (SVD) for randomized SVD
cpdef void memFinalSVD(
		const u8[:,::1] G, f32[:,::1] U, const f32[:,::1] V, f32[:,::1] X, f32[::1] f, f32[::1] d, const Py_ssize_t M_w
	) noexcept nogil:
	cdef:
		Py_ssize_t M = X.shape[0]
		Py_ssize_t N = X.shape[1]
		Py_ssize_t B = G.shape[1]
		Py_ssize_t K = U.shape[1]
		size_t b, i, j, l, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fl, dl
		f32* Ul
	for j in prange(M, schedule='guided'):
		i = 0
		l = M_w + j
		fl = f[l]
		dl = d[l]
		Ul = &U[l,0]
		for b in range(B):
			byte = G[l,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					X[j,i] = (<f32>g - 2.0*fl)*dl
				else:
					X[j,i] = innerE(Ul, &V[i,0], fl, K)*dl
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break



"""
EMU.
Main caller. Performs iterative SVD of allele count matrix (EM-PCA).

Jonas Meisner, Siyang Liu, Mingxi Huang and Anders Albrechtsen
"""

__author__ = "Jonas Meisner"

# Libraries
import argparse
import os
import sys
from datetime import datetime
from time import time

VERSION = "1.3.0"

# Argparse
parser = argparse.ArgumentParser(prog="emu")
parser.add_argument("--version", action="version", \
	version=f"{VERSION}")
parser.add_argument("-b", "--bfile", metavar="FILE-PREFIX",
	help="Prefix for PLINK files (.bed, .bim, .fam)")
parser.add_argument("-e", "--eig", metavar="INT", type=int,
	help="Number of eigenvectors to use in iterative estimation")
parser.add_argument("-t", "--threads", metavar="INT", type=int, default=1,
	help="Number of threads")
parser.add_argument("-f", "--maf", metavar="FLOAT", type=float,
	help="Threshold for minor allele frequencies")
parser.add_argument("-s", "--selection", action="store_true",
	help="Perform PC-based selection scan (Galinsky et al. 2016)")
parser.add_argument("-o", "--out", metavar="OUTPUT", default="emu",
	help="Prefix output name",)
parser.add_argument("-m", "--mem", action="store_true",
	help="Memory-efficient variant")
parser.add_argument("--iter", metavar="INT", type=int, default=100,
	help="Maximum iterations in estimation of individual allele frequencies (100)")
parser.add_argument("--tole", metavar="FLOAT", type=float, default=5e-7,
	help="Tolerance in update for individual allele frequencies (5e-7)")
parser.add_argument("--power", metavar="INT", type=int, default=10,
	help="Number of power iterations in randomized SVD (10)")
parser.add_argument("--batch", metavar="INT", type=int, default=8192,
	help="Number of SNPs to use in batches of memory variant (8192)")
parser.add_argument("--seed", metavar="INT", type=int, default=42,
	help="Set random seed")
parser.add_argument("--eig-out", metavar="INT", type=int,
	help="Number of eigenvectors to output in final SVD")
parser.add_argument("--loadings", action="store_true",
	help="Save SNP loadings")
parser.add_argument("--raw", action="store_true",
	help="Raw output without '*.fam' info")



##### EMU main caller #####
def main():
	args = parser.parse_args()
	if len(sys.argv) < 2:
		parser.print_help()
		sys.exit()
	print("--------------------------------")
	print(f"EMU v{VERSION}")
	print(f"Using {args.threads} thread(s).")
	print("--------------------------------\n")

	# Check input
	assert args.bfile is not None, "No input data (--bfile)"
	assert args.eig is not None, "Must specify number of eigenvectors to use!"
	assert args.threads > 0, "Please select a valid number of threads!"
	assert args.seed >= 0, "Please select a valid seed!"
	assert args.iter >= 0, "Please select a valid number of iterations!"
	assert args.tole >= 0.0, "Please select a valid tolerance!"
	assert args.batch > 1, "Please select a valid number of batches!"
	assert args.power > 1, "Please select a valid number of power iterations!"
	if args.eig_out is not None:
		assert args.eig_out > 1, "Please select a valid number of output eigenvectors!"
	start = time()

	# Create log-file of arguments
	full = vars(parser.parse_args())
	deaf = vars(parser.parse_args([]))
	with open(f"{args.out}.log", "w") as log:
		log.write(f"EMU v{VERSION}\n")
		log.write(f"Time: {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}\n")
		log.write(f"Directory: {os.getcwd()}\n")
		log.write("Options:\n")
		for key in full:
			if full[key] != deaf[key]:
				if type(full[key]) is bool:
					log.write(f"\t--{key}\n")
				else:
					log.write(f"\t--{key} {full[key]}\n")
	del full, deaf

	# Control threads of external numerical libraries
	os.environ["MKL_NUM_THREADS"] = str(args.threads)
	os.environ["MKL_MAX_THREADS"] = str(args.threads)
	os.environ["OMP_NUM_THREADS"] = str(args.threads)
	os.environ["OMP_MAX_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_NUM_THREADS"] = str(args.threads)
	os.environ["NUMEXPR_MAX_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_NUM_THREADS"] = str(args.threads)
	os.environ["OPENBLAS_MAX_THREADS"] = str(args.threads)

	# Load numerical libraries
	import numpy as np
	from emu import functions
	from emu import shared

	# Set K
	if args.eig_out is None:
		K = args.eig
	else:
		K = args.eig_out

	### Read data
	assert os.path.isfile(f"{args.bfile}.bed"), "bed file doesn't exist!"
	assert os.path.isfile(f"{args.bfile}.bim"), "bim file doesn't exist!"
	assert os.path.isfile(f"{args.bfile}.fam"), "fam file doesn't exist!"
	print("Reading data...", end="", flush=True)
	G, M, N = functions.readPlink(args.bfile)
	print(f"\rLoaded {N} samples and {M} SNPs.")

	# Estimate allele frequencies
	f = np.zeros(M, dtype=np.float32)
	d = np.zeros(M, dtype=np.float32)
	n = np.zeros(M, dtype=np.uint32)
	shared.estimateF(G, f, d, n, N)

	# Check input
	if np.allclose(n, np.full(M, N, dtype=np.uint32)):
		print("No missingness in data!")
		args.iter = 0
	assert (not np.allclose(np.max(f), 1.0)) or (not np.allclose(np.min(f), 0.0)), \
		"Fixed sites in dataset. Please perform MAF filtering!"
	del n

	### Perform EM-PCA
	print(f"\nPerforming EMU using {args.eig} eigenvector(s).")
	rng = np.random.default_rng(args.seed)
	if args.mem:
		print("Using memory-efficient variant of EMU.")
		U, S, V, it, converged = functions.emuAlgorithm(G, None, f, d, N, args.eig, K, \
			args.iter, args.tole, args.batch, args.power, rng)
	else:
		E = np.zeros((M, N), dtype=np.float32)
		U, S, V, it, converged = functions.emuAlgorithm(G, E, f, d, N, args.eig, K, 
			args.iter, args.tole, args.batch, args.power, rng)
		del E
	del G, f, d

	# Print elapsed time for estimation
	t_tot = time()-start
	t_min = int(t_tot//60)
	t_sec = int(t_tot - t_min*60)
	print(f"\nTotal elapsed time: {t_min}m{t_sec}s")

	# Save matrices
	if args.raw:
		np.savetxt(f"{args.out}.eigvecs", V, fmt="%.7f")
	else:
		F = np.loadtxt(f"{args.bfile}.fam", usecols=[0,1], dtype=np.str_)
		h = ["#FID", "IID"] + [f"PC{k}" for k in range(1, K+1)]
		V = np.hstack((F, np.round(V, 7)))
		np.savetxt(f"{args.out}.eigvecs", V, fmt="%s", delimiter="\t", header="\t".join(h), comments="")
	print(f"Saved eigenvector(s) as {args.out}.eigvecs")
	np.savetxt(f"{args.out}.eigvals", (S**2)/float(M), fmt="%.7f")
	print(f"Saved eigenvalue(s) as {args.out}.eigvals")
	del V, S

	# Save loadings
	if args.loadings:
		np.savetxt(f"{args.out}.loadings", U, fmt="%.7f")
		print(f"Saved SNP loadings as {args.out}.loadings")

	# Perform genome-wide selection scan
	if args.selection:
		Dsquared = np.zeros((M, K), dtype=np.float32)
		shared.galinskyScan(U, Dsquared)
		np.savetxt(f"{args.out}.selection", Dsquared, fmt="%.7f")
		print(f"Saved test statistics as {args.out}.selection")
		del Dsquared
	del U

	# Write output info to log-file
	with open(f"{args.out}.log", "a") as log:
		if args.iter > 0:
			if converged:
				log.write(f"\nEM-PCA converged in {it} iterations.\n")
			else:
				log.write("\nEM-PCA did not converge!\n")
		log.write(f"Total elapsed time: {t_min}m{t_sec}s\n")
		log.write(f"Saved eigenvalue(s) as {args.out}.eigvals\n")
		log.write(f"Saved eigenvector(s) as {args.out}.eigvecs\n")
		if args.loadings:
			log.write(f"Saved SNP loadings as {args.out}.loadings\n")
		if args.selection:
			log.write(f"Saved test statistics as {args.out}.selection\n")



##### Main exception #####
assert __name__ != "__main__", "Please use the 'emu' command!"



import numpy as np
from math import ceil
from time import time
from emu import memory
from emu import shared

##### EMU functions #####
### Read PLINK files
def readPlink(bfile):
	# Find length of fam-file
	N = 0
	with open(f"{bfile}.fam", "r") as fam:
		for _ in fam:
			N += 1
	N_bytes = ceil(N/4) # Length of bytes to describe N individuals

	# Read .bed file
	with open(f"{bfile}.bed", "rb") as bed:
		G = np.fromfile(bed, dtype=np.uint8, offset=3)
	assert (G.shape[0] % N_bytes) == 0, "bim file doesn't match!"
	M = G.shape[0]//N_bytes
	G.shape = (M, N_bytes)
	return G, M, N


### Helper function
# Flip signs of SVD output - Based on scikit-learn (svd_flip)
def signFlip(U, V):
    mcols = np.argmax(np.abs(U), axis=0)
    signs = np.sign(U[mcols, range(U.shape[1])])
    U *= signs
    V *= signs
    return U, V


### Randomized SVD functions
# SVD through eigendecomposition
def eigSVD(H):
	D, V = np.linalg.eigh(np.dot(H.T, H))
	S = np.sqrt(D)
	U = np.dot(H, V*(1.0/S))
	return np.ascontiguousarray(U[:,::-1]), np.ascontiguousarray(S[::-1]), np.ascontiguousarray(V[:,::-1])

# Randomized SVD with dynamic shift
def randomizedSVD(E, K, power, rng):
	M, N = E.shape
	a = 0.0
	L = max(K + 10, 20)
	H = np.zeros((N, L), dtype=np.float32)
	A = rng.standard_normal(size=(M, L), dtype=np.float32)

	# Prime iteration
	np.dot(E.T, A, out=H)
	Q, _, _ = eigSVD(H)

	# Power iterations
	for _ in np.arange(power):
		np.dot(E, Q, out=A)
		np.dot(E.T, A, out=H)
		H -= a*Q
		Q, S, _ = eigSVD(H)
		if S[-1] > a:
			a = 0.5*(S[-1] + a)
	
	# Extract singular vectors
	np.dot(E, Q, out=A)
	U, S, V = eigSVD(A)
	return np.ascontiguousarray(U[:,:K]), S[:K], np.ascontiguousarray(np.dot(Q, V)[:,:K])

# Batched randomized SVD with dynamic shift
def memorySVD(G, U0, V0, f, d, N, K, batch, power, rng):
	M = G.shape[0]
	W = ceil(M/batch)
	a = 0.0
	L = max(K + 10, 20)
	H = np.zeros((N, L), dtype=np.float32)
	X = np.zeros((batch, N), dtype=np.float32)
	A = rng.standard_normal(size=(M, L), dtype=np.float32)

	# Prime iteration
	for w in np.arange(W):
		M_w = w*batch
		if w == (W-1): # Last batch
			X = np.zeros((M - M_w, N), dtype=np.float32)
		if d is None:
			if U0 is None:
				memory.memCenter(G, X, f, M_w)
			else:
				memory.memCenterSVD(G, U0, V0, X, f, M_w)
		else:
			if U0 is None:
				memory.memFinal(G, X, f, d, M_w)
			else:
				memory.memFinalSVD(G, U0, V0, X, f, d, M_w)
		H += np.dot(X.T, A[M_w:(M_w + X.shape[0])])
	Q, _, _ = eigSVD(H)
	H.fill(0.0)

	# Power iterations
	for _ in np.arange(power):
		X = np.zeros((batch, N), dtype=np.float32)
		for w in np.arange(W):
			M_w = w*batch
			if w == (W-1): # Last batch
				X = np.zeros((M - M_w, N), dtype=np.float32)
			if d is None:
				if U0 is None:
					memory.memCenter(G, X, f, M_w)
				else:
					memory.memCenterSVD(G, U0, V0, X, f, M_w)
			else:
				if U0 is None:
					memory.memFinal(G, X, f, d, M_w)
				else:
					memory.memFinalSVD(G, U0, V0, X, f, d, M_w)
			A[M_w:(M_w + X.shape[0])] = np.dot(X, Q)
			H += np.dot(X.T, A[M_w:(M_w + X.shape[0])])
		H -= a*Q
		Q, S, _ = eigSVD(H)
		H.fill(0.0)
		if S[-1] > a:
			a = 0.5*(S[-1] + a)

	# Extract singular vectors
	X = np.zeros((batch, N), dtype=np.float32)
	for w in np.arange(W):
		M_w = w*batch
		if w == (W-1): # Last batch
			X = np.zeros((M - M_w, N), dtype=np.float32)
		if d is None:
			if U0 is None:
				memory.memCenter(G, X, f, M_w)
			else:
				memory.memCenterSVD(G, U0, V0, X, f, M_w)
		else:
			if U0 is None:
				memory.memFinal(G, X, f, d, M_w)
			else:
				memory.memFinalSVD(G, U0, V0, X, f, d, M_w)
		A[M_w:(M_w + X.shape[0])] = np.dot(X, Q)
	U, S, V = eigSVD(A)
	return np.ascontiguousarray(U[:,:K]), S[:K], np.ascontiguousarray(np.dot(Q, V)[:,:K])


### EMU algorithm
def emuAlgorithm(G, E, f, d, N, e, K, iter, tole, batch, power, rng):
	M = G.shape[0]

	# Exit without performing EMU
	if iter < 1:
		print("Warning, no EM-PCA iterations are performed!")
		print(f"Extracting {K} eigenvector(s).")
		if E is None:
			U, S, V = memorySVD(G, None, None, f, d, N, K, batch, power, rng)
		else:
			shared.standardInit(G, E, f, d)
			U, S, V = randomizedSVD(E, K, power, rng)
		U, V = signFlip(U, V)
		return U, S, V, 0, False
	else:
		# Set up acceleration containers
		U1 = np.zeros((M, e), dtype=np.float32)
		U2 = np.zeros((M, e), dtype=np.float32)
		V1 = np.zeros((N, e), dtype=np.float32)
		V2 = np.zeros((N, e), dtype=np.float32)

		# Estimate initial individual allele frequencies
		print("Initiating accelerated EM scheme")
		if E is None:
			U, S, V = memorySVD(G, None, None, f, None, N, e, batch, power, rng)
		else:
			shared.centerInit(G, E, f)
			U, S, V = randomizedSVD(E, K, power, rng)
		U, V = signFlip(U, V)
		V *= S
		U_pre = np.copy(U)

		# Iterative estimation of individual allele frequencies
		ts = time()
		for it in range(1, iter+1):
			# 1st SVD step
			if E is None:
				U1, S1, V1 = memorySVD(G, U, V, f, None, N, e, batch, power, rng)
			else:
				shared.centerAccel(G, E, U, V, f)
				U1, S1, V1 = randomizedSVD(E, e, power, rng)
			U1, V1 = signFlip(U1, V1)
			V1 *= S1

			# 2nd SVD step
			if E is None:
				U2, S2, V2 = memorySVD(G, U1, V1, f, None, N, e, batch, power, rng)
			else:
				shared.centerAccel(G, E, U1, V1, f)
				U2, S2, V2 = randomizedSVD(E, e, power, rng)
			U2, V2 = signFlip(U2, V2)
			V2 *= S2

			# QN steps
			shared.alphaStep(U, U1, U2)
			shared.alphaStep(V, V1, V2)

			# Stabilization step
			if E is None:
				U, S, V = memorySVD(G, U, V, f, None, N, e, batch, power, rng)
			else:
				shared.centerAccel(G, E, U, V, f)
				U, S, V = randomizedSVD(E, e, power, rng)
			U, V = signFlip(U, V)
			V *= S

			# Break iterative update if converged
			rmseU = shared.rmse(U, U_pre)
			print(f"({it})\tRMSE = {rmseU:.8f}\t({time()-ts:.1f}s)")
			if rmseU < tole:
				print("EM-PCA has converged.")
				converged = True
				break
			if it == iter:
				print("EM-PCA did not converge!")
				converged = False
			memoryview(U_pre.ravel())[:] = memoryview(U.ravel())
			ts = time()
		del U1, U2, U_pre, V1, V2, S1, S2

		# Estimating final SVD
		print(f"Extracting {K} eigenvector(s).")
		if E is None:
			U, S, V = memorySVD(G, U, V, f, d, N, K, batch, power, rng)
		else:
			shared.standardAccel(G, E, U, V, f, d)
			U, S, V = randomizedSVD(E, K, power, rng)
		U, V = signFlip(U, V)
		return U, S, V, it, converged
	


# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True
import numpy as np
cimport numpy as np
from cython.parallel import prange
from libc.math cimport sqrtf
from libc.stdint cimport uint8_t, uint32_t

ctypedef uint8_t u8
ctypedef uint32_t u32
ctypedef float f32

cdef f32 PRO_MIN = 1e-4
cdef f32 PRO_MAX = 2.0 - (1e-4)
cdef f32 ACC_MIN = 1.0
cdef f32 ACC_MAX = 256.0
cdef inline f32 _fmax(f32 a, f32 b) noexcept nogil: return a if a > b else b
cdef inline f32 _fmin(f32 a, f32 b) noexcept nogil: return a if a < b else b
cdef inline f32 _clamp1(f32 a) noexcept nogil: return _fmax(PRO_MIN, _fmin(a, PRO_MAX))
cdef inline f32 _clamp2(f32 a) noexcept nogil: return _fmax(ACC_MIN, _fmin(a, ACC_MAX))


##### EMU-mem #####
# Inline functions
cdef inline f32 innerE(
		const f32* u, const f32* v, const f32 f, const Py_ssize_t K
	) noexcept nogil:
	cdef:
		size_t k
		f32 d = 2.0*f
		f32 e = d
	for k in range(K):
		e += u[k]*v[k]
	return _clamp1(e) - d

cdef inline f32 computeC(
		const f32* x0, const f32* x1, const f32* x2, const Py_ssize_t I
	) noexcept nogil:
	cdef:
		size_t i
		f32 sum1 = 0.0
		f32 sum2 = 0.0
		f32 c, u, v
	for i in prange(I):
		u = x1[i] - x0[i]
		v = x2[i] - x1[i] - u
		sum1 += u*u
		sum2 += u*v
	c = -(sum1/sum2)
	return _clamp2(c)

cdef inline void updateAlpha(
		f32* x0, const f32* x1, const f32* x2, const f32 c1, const Py_ssize_t I
	) noexcept nogil:
	cdef:
		size_t i
		f32 c2 = 1.0 - c1
	for i in prange(I):
		x0[i] = c2*x1[i] + c1*x2[i]

# Estimate population allele frequencies
cpdef void estimateF(
		const u8[:,::1] G, f32[::1] f, f32[::1] d, u32[::1] n, const Py_ssize_t N
	) noexcept nogil:
	cdef:
		Py_ssize_t M = G.shape[0]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
	for j in prange(M, schedule='guided'):
		i = 0
		for b in range(B):
			byte = G[j,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					f[j] += <f32>g
					n[j] += 1
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break
		if n[j] > 0:
			f[j] /= <f32>(2*n[j])
			if (f[j] > 0.0) and (f[j] < 1.0):
				d[j] = 1.0/sqrtf(2.0*f[j]*(1.0 - f[j]))
		else:
			f[j] = 0.0

# Initialize and standardize E
cpdef void standardInit(
		const u8[:,::1] G, f32[:,::1] E, f32[::1] f, f32[::1] d
	) noexcept nogil:
	cdef:
		Py_ssize_t M = E.shape[0]
		Py_ssize_t N = E.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fj, dj
	for j in prange(M, schedule='guided'):
		i = 0
		fj = f[j]
		dj = d[j]
		for b in range(B):
			byte = G[j,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					E[j,i] = (<f32>g - 2.0*fj)*dj
				else:
					E[j,i] = 0.0
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Initialize and center E
cpdef void centerInit(
		const u8[:,::1] G, f32[:,::1] E, f32[::1] f
	) noexcept nogil:
	cdef:
		Py_ssize_t M = E.shape[0]
		Py_ssize_t N = E.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fj
	for j in prange(M, schedule='guided'):
		i = 0
		fj = f[j]
		for b in range(B):
			byte = G[j,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					E[j,i] = <f32>g - 2.0*fj
				else:
					E[j,i] = 0.0
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Standardize E in acceleration scheme
cpdef void standardAccel(
		const u8[:,::1] G, f32[:,::1] E, f32[:,::1] U, f32[:,::1] V, f32[::1] f, f32[::1] d
	) noexcept nogil:
	cdef:
		Py_ssize_t M = E.shape[0]
		Py_ssize_t N = E.shape[1]
		Py_ssize_t K = U.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fj, dj
		f32* Uj
	for j in prange(M, schedule='guided'):
		i = 0
		fj = f[j]
		dj = d[j]
		Uj = &U[j,0]
		for b in range(B):
			byte = G[j,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					E[j,i] = (<f32>g - 2.0*fj)*dj
				else:
					E[j,i] = innerE(Uj, &V[i,0], fj, K)*dj
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Center E in acceleration scheme
cpdef void centerAccel(
		const u8[:,::1] G, f32[:,::1] E, f32[:,::1] U, f32[:,::1] V, f32[::1] f
	) noexcept nogil:
	cdef:
		Py_ssize_t M = E.shape[0]
		Py_ssize_t N = E.shape[1]
		Py_ssize_t K = U.shape[1]
		Py_ssize_t B = G.shape[1]
		size_t b, i, j, bytepart
		u8[4] recode = [2, 9, 1, 0]
		u8 mask = 3
		u8 g, byte
		f32 fj
		f32* Uj
	for j in prange(M, schedule='guided'):
		i = 0
		fj = f[j]
		Uj = &U[j,0]
		for b in range(B):
			byte = G[j,b]
			for bytepart in range(4):
				g = recode[byte & mask]
				if g != 9:
					E[j,i] = <f32>g - 2.0*fj
				else:
					E[j,i] = innerE(Uj, &V[i,0], fj, K)
				byte = byte >> 2 # Right shift 2 bits
				i = i + 1
				if i == N:
					break

# Accelerated QN jump
cpdef void alphaStep(
		f32[:,::1] X0, const f32[:,::1] X1, const f32[:,::1] X2
	) noexcept nogil:
	cdef:
		Py_ssize_t I = X0.shape[0]
		Py_ssize_t J = X0.shape[1]
		f32 c
	c = computeC(&X0[0,0], &X1[0,0], &X2[0,0], I*J)
	updateAlpha(&X0[0,0], &X1[0,0], &X2[0,0], c, I*J)

# Root-mean squared error
cpdef f32 rmse(
		const f32[:,::1] A, const f32[:,::1] B
	) noexcept nogil:
	cdef:
		Py_ssize_t M = A.shape[0]
		Py_ssize_t K = A.shape[1]
		size_t j, k
		f32 res = 0.0
	for j in prange(M, schedule='guided'):
		for k in range(K):
			res += (A[j,k] - B[j,k])*(A[j,k] - B[j,k])
	return sqrtf(res/(<f32>(M*K)))

# Selection scan
cpdef void galinskyScan(
		const f32[:,::1] U, f32[:,::1] Dsquared
	) noexcept nogil:
	cdef:
		Py_ssize_t M = U.shape[0]
		Py_ssize_t K = U.shape[1]
		size_t j, k
		f32 m = <f32>M
	for j in prange(M, schedule='guided'):
		for k in range(K):
			Dsquared[j,k] = (U[j,k]*U[j,k])*m
